##
# fish.rb - 2006
# 
# This is a Ruby implementation which allows you do encrypt/decrypt messages
# generated by the popular FiSH and Mircryption scripts on IRC. You can use
# it to enable your Ruby IRC bot in blowfish-channels for example.
# 
# = Usage
# 
#   fish = IRC::FiSH.new(key)
#   encrypted_line = fish.encrypt(line)
#   decrypted_line = fish.decrypt(encrypted_line)
# 
#   # now we have: line == decrypted_line
#
# Note that this code does not strip a line from the prefix such as
# "+OK" and "mcps". You will still need to do that part yourself.
# 
# = Thanks
# 
# Thanks to the people at Mircryption for putting example implementations
# on their website.
#
##

require 'crypt/blowfish'

module IRC

# error class for input we cannot handle
class BadInputError < StandardError; end

# non-standard implementation of Base64 encoding/decoding which is
# compatible with FiSH, Mircryption.
class MBase64
  B64 = './0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';

  # decode (sub)string of multiple of 12 characters
  def self.decode(encoded)
    # verify that the input is OK
    if not encoded.length % 12 == 0
      raise IRC::BadInputError,
        "can only decode strings which are a multiple of 12 characters."
    end
    
    decoded = String.new

    k = -1
    while (k < encoded.length - 1) do
      right = 0
      left = 0
  
      (0..5).each do |i|
        k = k + 1
        right |= B64.index(encoded[k]) << (i * 6)
      end

      (0..5).each do |i|
        k = k + 1
        left |= B64.index(encoded[k]) << (i * 6)
      end
      
      (0..3).each do |i|
        decoded += ((left & (0xFF << ((3 - i) * 8))) >> ((3 - i) * 8)).chr
      end
      
      (0..3).each do |i|
        decoded += ((right & (0xFF << ((3 - i) * 8))) >> ((3 - i) * 8)).chr
      end
    end # while

    return decoded
  end # def self.decode(encoded)
  
  # encode string of a multiple of 8 characters into encoded
  # string which is a multiple of 12 characters..
  def self.encode(decoded)
    # verify that the input is OK
    if not decoded.length % 8 == 0
      raise IRC::BadInputError,
        "can only encode strings which are a multiple of 8 characters."
    end
    
    encoded = String.new

    k = -1

    while (k < decoded.length - 1) do
      k = k.next
      left = (decoded[k] << 24)
      k = k.next
      left += (decoded[k] << 16)
      k = k.next
      left += (decoded[k] << 8)
      k = k.next
      left += decoded[k]

      k = k.next
      right = (decoded[k] << 24)
      k = k.next
      right += (decoded[k] << 16)
      k = k.next
      right += (decoded[k] << 8)
      k = k.next
      right += decoded[k]

      (0..5).each do
        encoded += B64[right & 0x3F].chr
        right = right >> 6
      end
      (0..5).each do
        encoded += B64[left & 0x3F].chr
        left = left >> 6
      end
    end # while
    
    return encoded
  end
  
end # class MBase64

# class for encrypting/decrypting lines of text compatible with
# FiSH/Mircryption for irc
class FiSH
  def initialize(key)
    @blowfish = Crypt::Blowfish.new(key)
  end

  # encrypt text and return encrypted string
  def encrypt(text)
    text = pad(text, 8)
    result = ""
    
    # for every block of 8 in text, encode with blowfish ECB
    # and do base64 encoding the mirc way
    num_block = text.length / 8
    num_block.times do |n|
      block = text[n*8..(n+1)*8-1]
      enc = @blowfish.encrypt_block(block)
      result += MBase64.encode(enc)
    end

    return result
  end

  # decrypt text and return decoded string
  def decrypt(text)
    # verify that the input is OK
    if not text.length % 12 == 0
      raise IRC::BadInputError,
        "can only decrypt strings which are a multiple of 12 characters."
    end
    
    result = ""

    # for every block of 12 characters
    num_block = (text.length / 12).to_i
    num_block.times do |n|
      block = MBase64.decode( text[n*12..(n+1)*12-1] )
      result += @blowfish.decrypt_block(block)
    end

    # strip padding off the result
    return result.gsub(/\0*$/, '')
  end
  
  private

  # make sure text is a multiple of n by padding it with \0's
  def pad(text, n=8)
    pad_num = n - (text.length % n)
    if pad_num > 0 and pad_num != n
      pad_num.times { text += 0.chr }
    end
    return text
  end
end

end # module IRC


# Example usage
if __FILE__ == $0
  key='tQ0V3Ca7eT5bjnwWoT4GBVJjstpqygKC3uzThhdKUWFb91YyYxPXeW7T'
  test = "Say something loud! Maybe he'll use it in his examples."

  fish = IRC::FiSH.new(key)

  puts "plain  : " + test
  
  enc = fish.encrypt(test)
  puts "encoded: " + enc
  
  dec = fish.decrypt(enc)
  puts "decoded: " + dec

  exit
end
